/*
    smix - audio mixer
   
    This version Copyright (C) 1997 by Sven Anders

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <forms.h>
#include "mybutton.h"
#include "smix.xpm"

#include <Xutil.h>
#include <xpm.h>

#include <malloc.h>
#include <signal.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

/*----------------------------------------------------------------------------------*/

/* Name of the program */

#ifndef PRG_NAME
#define PRG_NAME  "smix"
#endif

/*----------------------------------------------------------------------------------*/

/* Name of config file */
/* A ~ at the beginning will be subsituted with home directory. */

#ifndef CONFIGFILE
  char *config_file = "~/." PRG_NAME ".conf";
#else
  char *config_file = CONFIGFILE;
#endif

/*----------------------------------------------------------------------------------*/

/* Name of mixer device */

#ifndef DEFAULT_MIXERDEVICE
  char *dev_name = "/dev/mixer";
#else
  char *dev_name = DEFAULT_MIXERDEVICE;
#endif

/*----------------------------------------------------------------------------------*/

/* colors */

#define LEFT_COLOR     FL_WHITE
#define RIGHT_COLOR    FL_RED
#define MONO_COLOR     FL_BLACK
#define BALANCE_COLOR  FL_GREEN
#define REC_COLOR      FL_RED
#define LOCK_COLOR     FL_BLACK
#define MUTE_COLOR     FL_BLUE
#define SOLO_COLOR     FL_CYAN
#define INACTIVE_COLOR FL_INACTIVE

/*----------------------------------------------------------------------------------*/

#define MAX_NUM_MIXER_DEVS  20
#define MAX_MIXER_VALUE     100

char *dev_labels[] = SOUND_DEVICE_LABELS;
struct mixer_info  mixerinfo;

struct mixer
{
  struct mixer_info info;
  int handle;
  int devmask;
  int stereomask;
  int recmask;
  int recsrc;

  int devs;
  int rec_devs;
  int stereo_devs;
} mix;

struct mix_channel
{
  char *label;               /* the channel's name */

  int left;                  /* device values */
  int right;
  int stereo;
  int dev_num;
  int dev_bit;
  int rec_src;
  int rec_support;

  int muted_l;               /* saved muted values */
  int muted_r;

  int show;                  /* show channel ? */
  int show_balance;          /* show balance slider ? */
  int solo_channel;          /* solo enabled channel ? */

  int button_mute;           /* buttons pressed ? */
  int button_lock;
  int button_solo;

  int saved_button_mute;     /* saved 'button_mute' */
  int saved_button_solo;     /* saved 'button_solo' */

  FL_OBJECT *fl_l, *fl_r;    /* xforms objects */
  FL_OBJECT *fl_bal;
  FL_OBJECT *fl_rec;
  FL_OBJECT *fl_mute;
  FL_OBJECT *fl_lock;
  FL_OBJECT *fl_solo;

} mchannel[MAX_NUM_MIXER_DEVS];

int button_global_mute = FALSE;
int button_mute_all = FALSE;
int button_lock_all = FALSE;
int button_rec_all = FALSE;

int show_overall_balance = TRUE;
int alternate_appearance = FALSE;
int exclusive_rec_selection = TRUE;
int muting_enabled = TRUE;
int locking_enabled = TRUE;
int recording_enabled = TRUE;
int active_update = TRUE;

/*----------------------------------------------------------------------------------*/

FL_FORM  *mix_form = NULL;
FL_FORM  *config_form = NULL;
FL_FORM  *help_form = NULL;
FL_OBJECT *help_button = NULL;
FL_OBJECT *config_button = NULL;
FL_OBJECT *overall_balance = NULL;
FL_OBJECT *rec_all = NULL;
FL_OBJECT *mute_all = NULL;
FL_OBJECT *lock_all = NULL;
FL_OBJECT *mute_global = NULL;

/*----------------------------------------------------------------------------------*/

int verbose = FALSE;

/*----------------------------------------------------------------------------------*/

int sig_alarm = FALSE;
int callback_depth = 0;
static struct sigaction sig_alarm_action;
static struct sigaction old_sig_alarm_action;
static struct itimerval sig_alarm_delay= { {0, 33333}, {0, 33333} };
static struct itimerval old_sig_alarm_delay;

/*----------------------------------------------------------------------------------*/
/* xforms independent functions */
/*----------------------------------------------------------------------------------*/

inline const int max( int a, int b)
{
    return ( a < b ) ? b : a;
}

inline const int min( int a, int b)
{
    return ( a > b ) ? b : a;
}

char *err_msg[]={ "Cannot open mixer",
		  "Error reading from mixer",
		  "No mixer devices found",
		  "Error writing to mixer"
		};

void end_error(int err)
{
  fprintf(stderr, "%s\n", err_msg[err-1]);
  exit(err);
}

void set_volume(int channel, int left, int right)
{
  unsigned char temp[4];

  mchannel[channel].left = left;
  mchannel[channel].right = right;
  temp[0]=(unsigned char)left;
  temp[1]=(unsigned char)right;
  temp[2]=temp[3]=0;
  ioctl(mix.handle, MIXER_WRITE(mchannel[channel].dev_num), temp);
}

void get_volume(int channel, int *left, int *right)
{
  unsigned char temp[4];

  ioctl(mix.handle, MIXER_READ(mchannel[channel].dev_num), &temp);
  (unsigned char) *left = temp[0];
  (unsigned char) *right = temp[1];
}

void init_channel_struct(void)
{
  unsigned int i,j;

  if ((mix.handle=open(dev_name, O_RDWR))<0) end_error(1);

  if (ioctl(mix.handle, SOUND_MIXER_INFO, &mix.info)==-1) end_error(2);
  if (ioctl(mix.handle, SOUND_MIXER_READ_DEVMASK, &mix.devmask)==-1) end_error(2);
  if (ioctl(mix.handle, SOUND_MIXER_READ_STEREODEVS, &mix.stereomask)==-1) end_error(2);
  if (ioctl(mix.handle, SOUND_MIXER_READ_RECMASK, &mix.recmask)==-1) end_error(2);
  if (ioctl(mix.handle, SOUND_MIXER_READ_RECSRC, &mix.recsrc)==-1) end_error(2);

  if (!mix.devmask) end_error(3);

  mix.devs = 0;
  mix.rec_devs = 0;
  mix.stereo_devs = 0;

  for (i=0, j=1; i<MAX_NUM_MIXER_DEVS; i++, j<<=1)
  {
    if (mix.devmask & j)
    {
      mchannel[mix.devs].label = (char *)malloc(strlen(dev_labels[i])+1);
      strcpy(mchannel[mix.devs].label, dev_labels[i]);

      if ((mix.stereomask & j) != 0)
      {
	mchannel[mix.devs].stereo = TRUE;
	mix.stereo_devs++;
      }
      else
	mchannel[mix.devs].stereo = FALSE;

      if ((mix.recmask & j) != 0)
      {
	mchannel[mix.devs].rec_support = TRUE;
	mix.rec_devs++;
      }
      else
	mchannel[mix.devs].rec_support = FALSE;

      if ((mix.recsrc & j) != 0)
	mchannel[mix.devs].rec_src = TRUE;
      else
	mchannel[mix.devs].rec_src = FALSE;

      mchannel[mix.devs].dev_num = i;
      mchannel[mix.devs].dev_bit = j;
      get_volume(mix.devs, &mchannel[mix.devs].left, &mchannel[mix.devs].right);

      mchannel[mix.devs].muted_l = -1;
      mchannel[mix.devs].muted_r = -1;

      mchannel[mix.devs].show = TRUE;
      mchannel[mix.devs].show_balance = TRUE;
      mchannel[mix.devs].solo_channel = FALSE;

      mchannel[mix.devs].button_mute = FALSE;
      mchannel[mix.devs].button_lock = FALSE;
      mchannel[mix.devs].button_solo = FALSE;

      mchannel[mix.devs].saved_button_mute = FALSE;
      mchannel[mix.devs].saved_button_solo = FALSE;

      mchannel[mix.devs].fl_r = NULL;
      mchannel[mix.devs].fl_l = NULL;
      mchannel[mix.devs].fl_bal = NULL;
      mchannel[mix.devs].fl_rec = NULL;
      mchannel[mix.devs].fl_mute = NULL;
      mchannel[mix.devs].fl_lock = NULL;

      mix.devs++;
    }
  }
}

void print_value(int i)
{
  printf("%2u: %-7s: ", i, mchannel[i].label);

  if (mchannel[i].stereo)
    printf("stereo %u/%u", mchannel[i].left, mchannel[i].right);
  else
    printf("mono %u     ", mchannel[i].left);

  if (mchannel[i].rec_support) printf(" record supported");
  if (mchannel[i].rec_src) printf(", recording active");

  printf("\n");
}

void print_binary(int i)
{
  int j;
  for (j=0; j<32; j++)
  {
   if (i & (1 << j)) printf("1"); else printf("0");
   if (j == 19) printf(":");
  }
  printf("\n");
}

void print_values()
{
  int i;
  if (verbose)
  {
    printf("mixer id:   %s\n", mix.info.id);
    printf("mixer name: %s\n", mix.info.name);

    printf("mixer handle: %X\n", mix.handle);
    printf("mixer device mask:        "); print_binary(mix.devmask);
    printf("mixer stereo mask:        "); print_binary(mix.stereomask);
    printf("mixer recording mask:     "); print_binary(mix.recmask);
    printf("mixer rec. sources mask:  "); print_binary(mix.recsrc);

    printf("number of devices:        %u\n", mix.devs);
    printf("number of stereo devices: %u\n", mix.stereo_devs);
    printf("number of record devices: %u\n", mix.rec_devs);
    printf("\n");
  }
  for (i=0; i<mix.devs; i++) print_value(i);
  printf("\n");
}

void un_lock_channel(int channel)
{
  mchannel[channel].button_lock = ! mchannel[channel].button_lock;
}

void un_mute_channel(int channel)
{
  int i;

  if (mchannel[channel].button_mute)
  {
    mchannel[channel].button_mute = FALSE;
    mchannel[channel].left = mchannel[channel].muted_l;
    mchannel[channel].right = mchannel[channel].muted_r;
  }
  else
  {
    mchannel[channel].button_mute = TRUE;
    mchannel[channel].muted_l = mchannel[channel].left; mchannel[channel].left = 0;
    mchannel[channel].muted_r = mchannel[channel].right; mchannel[channel].right = 0;
  }

  if (mchannel[channel].solo_channel)
  {
    for (i=0; i<mix.devs; i++) mchannel[i].button_solo = FALSE;
  }

  set_volume(channel, mchannel[channel].left, mchannel[channel].right);
}

void un_solo_channel(int channel)
{
  int i;

  if (mchannel[channel].button_solo)
  {
    for (i=0; i<mix.devs; i++)
    {
      if ((mchannel[i].solo_channel) && (mchannel[i].button_mute))
      {
	un_mute_channel(i);
	mchannel[i].button_solo = FALSE;
      }
    }  
  }
  else
  {
    for (i=0; i<mix.devs; i++)
    {
      if (mchannel[i].solo_channel)
      {
	if (i == channel)
	{ 
	  if (mchannel[i].button_mute) un_mute_channel(i);
	}
	else
	{ 
	  if (!mchannel[i].button_mute) un_mute_channel(i);
	  mchannel[i].button_solo = FALSE;
	}
      }
    }
    mchannel[channel].button_solo = TRUE;
  }
}

void un_rec_channel(int channel)
{
  int i;

  if (exclusive_rec_selection)
  {
    for (i=0; i<mix.devs; i++)
    {
      if (i == channel) 
      {
	mchannel[channel].rec_src = TRUE;
	mix.recsrc |= mchannel[channel].dev_bit;	
      }
      else
      {
	mchannel[channel].rec_src = FALSE;
	mix.recsrc &= (0xffffffff ^ mchannel[i].dev_bit);
      }
    }
  }
  else
  {
    if (mchannel[channel].rec_src)
    {
      mchannel[channel].rec_src = FALSE;
      mix.recsrc &= (0xffffffff ^ mchannel[channel].dev_bit);
    }
    else
    {
      mchannel[channel].rec_src = TRUE;
      mix.recsrc |= mchannel[channel].dev_bit;
    }
  }

  if (ioctl(mix.handle, SOUND_MIXER_WRITE_RECSRC, &mix.recsrc)==-1) end_error(4);
}

void un_mute_all(void)
{
  int i;

  button_mute_all = ! button_mute_all;
  for (i=0; i<mix.devs; i++)
  {
    if ((mchannel[i].show) &&
        (mchannel[i].button_mute != button_mute_all)) un_mute_channel(i);
  }
}

void un_lock_all(void)
{
  int i;

  button_lock_all = ! button_lock_all;
  for (i=0; i<mix.devs; i++)
  {
    if ((mchannel[i].show) &&
        (mchannel[i].button_lock != button_lock_all)) un_lock_channel(i);
  }
}

void un_rec_all(void)
{
  int i;

  button_rec_all = ! button_rec_all;
  for (i=0; i<mix.devs; i++)
  {
    if ((mchannel[i].show) &&
        (mchannel[i].rec_src != button_rec_all)) un_rec_channel(i);
  }
}

void un_mute_global(void)
{
  int i;

  button_global_mute = ! button_global_mute;
  for (i=0; i<mix.devs; i++)
  {
    if (button_global_mute)
    {
      mchannel[i].saved_button_mute =  mchannel[i].button_mute;
      mchannel[i].saved_button_solo =  mchannel[i].button_solo;
      if (mchannel[i].button_mute != button_global_mute) un_mute_channel(i);
    }
    else
    {
      if (!mchannel[i].saved_button_mute) un_mute_channel(i);
      /* (re)set solo buttons */
      if (mchannel[i].saved_button_solo) un_solo_channel(i);
    }
  }
}

/*----------------------------------------------------------------------------------*/
/* xforms functions */
/*----------------------------------------------------------------------------------*/

void my_msleep(unsigned long delay)
{
  sigaction(SIGALRM, &old_sig_alarm_action, 0);
  setitimer(ITIMER_REAL, &old_sig_alarm_delay, 0);
  fl_msleep(delay);
  sigaction(SIGALRM, &sig_alarm_action, 0);
  setitimer(ITIMER_REAL, &sig_alarm_delay, 0);
}

/* update rec buttons in form */
void update_mixer_rec()
{
  int i;
  int recsrc_counter, showed_rec_counter;

  recsrc_counter = 0;
  showed_rec_counter = 0;

  if (ioctl(mix.handle, SOUND_MIXER_READ_RECSRC, &mix.recsrc) == -1) end_error(2);

  for (i=0; i<mix.devs; i++)
  {
    if ((mchannel[i].show) && (mchannel[i].rec_support))
    {
      showed_rec_counter++;
      if (mix.recsrc & mchannel[i].dev_bit)
      {
        recsrc_counter++;
	mchannel[i].rec_src = TRUE;
	if ((mchannel[i].fl_rec != NULL) && (fl_get_button(mchannel[i].fl_rec) == 0))
	    fl_set_button(mchannel[i].fl_rec, 1);
      }
      else
      {
	mchannel[i].rec_src = FALSE;
	if ((mchannel[i].fl_rec != NULL) && (fl_get_button(mchannel[i].fl_rec) == 1))
	    fl_set_button(mchannel[i].fl_rec, 0);
      }
    }
  }

  /* set rec_all button if necessary */
  if (rec_all != NULL) 
  {
    if ( button_rec_all != (recsrc_counter == showed_rec_counter)) un_rec_all();
    fl_set_button(rec_all, button_rec_all);
  }
}

/* update all mixer values in form */
void update_mixer_values()
{
  int i;
  int left, right;
  int bleft, bright, balance;
  int showed_lock_counter, 
      showed_silent_counter,
      global_silent_counter,
      showed_stereo_counter,
      show_counter;

  showed_stereo_counter = 0;
  showed_silent_counter = 0;
  global_silent_counter = 0;
  showed_lock_counter = 0;
  show_counter = 0;

  callback_depth++;
  fl_freeze_form(mix_form);
  update_mixer_rec();


  for (i=0; i<mix.devs; i++)
  {
    get_volume(i, &left, &right);

    /* has other mixer muted or unmuted? */
    if ((left == 0) && (right == 0))
    {
      if ((mchannel[i].left != 0) && (mchannel[i].right != 0) && 
	  (!mchannel[i].button_mute))
      {
	un_mute_channel(i);
	if (mchannel[i].fl_mute != NULL) fl_set_button(mchannel[i].fl_mute, 1);
      }
      global_silent_counter++;
      if (mchannel[i].show) showed_silent_counter++;
    }
    else
    {
      if (mchannel[i].button_mute)
      {
	un_mute_channel(i);
	if (mchannel[i].fl_mute != NULL) fl_set_button(mchannel[i].fl_mute, 0);
      }
    }
 
    if (mchannel[i].show)
    {
      show_counter++;

      if (!mchannel[i].stereo)
      {
	/* mono channel */
	if (mchannel[i].fl_l != NULL)
	  fl_set_object_color(mchannel[i].fl_l, FL_COL1,  MONO_COLOR);
      }
      else
      {
	/* stereo */
	showed_stereo_counter++;

	if (mchannel[i].button_lock)
	{
	  showed_lock_counter++;

	  /* 'lock' button pressed. */
	  if (mchannel[i].fl_bal != NULL)
	  {
	    //fl_set_slider_value(mchannel[i].fl_bal, 0);
	    fl_deactivate_object(mchannel[i].fl_bal);
	    fl_set_object_color(mchannel[i].fl_bal, FL_COL1,  INACTIVE_COLOR);
	  }

	  if (mchannel[i].fl_l != NULL)
            fl_set_object_color(mchannel[i].fl_l, FL_COL1,  LOCK_COLOR);
	  if (mchannel[i].fl_r != NULL)
	    fl_set_object_color(mchannel[i].fl_r, FL_COL1, LOCK_COLOR);
	}
	else
	{
	  /* 'lock' button not pressed. */
	  if (mchannel[i].fl_l != NULL)
	    fl_set_object_color(mchannel[i].fl_l, FL_COL1, LEFT_COLOR);
	  if (mchannel[i].fl_r != NULL)
	    fl_set_object_color(mchannel[i].fl_r, FL_COL1, RIGHT_COLOR);

	  if (mchannel[i].fl_bal != NULL)
	  {
	    fl_activate_object(mchannel[i].fl_bal);
	    fl_set_object_color(mchannel[i].fl_bal, FL_COL1,  BALANCE_COLOR);
	  }
	}
      }

      /* has volume changed ? */
      if (fl_get_slider_value(mchannel[i].fl_l) != left)
      {
	mchannel[i].left = left;
	fl_set_slider_value(mchannel[i].fl_l, left);
      }

      if (mchannel[i].stereo)
      {
	if (fl_get_slider_value(mchannel[i].fl_r) != right)
	{
	  mchannel[i].right = right;
	  fl_set_slider_value(mchannel[i].fl_r, right);
	}

	if (mchannel[i].fl_bal != NULL)
	{
	  if (left == right)
	    balance = 0;
	  else
	  {
	    if (left > right)
	    {
	      if (right == 0)
		balance = -MAX_MIXER_VALUE;
	      else
		balance = -MAX_MIXER_VALUE+(right*MAX_MIXER_VALUE/left);
	    }
	    else
	    {
	      if (left == 0)
		balance = MAX_MIXER_VALUE;
	      else
		balance = MAX_MIXER_VALUE-(left*MAX_MIXER_VALUE/right);
	    }
	  }
	  fl_set_slider_value(mchannel[i].fl_bal, balance);
	}
	else
	  mchannel[i].right = right;
      }
    }
  }

  /* set overall balance slider */
  if ((show_overall_balance) && (mix.stereo_devs > 0))
  {
    bleft = 0;
    bright = 0;
    balance = 0;
    for (i=0; i<mix.devs; i++)
    {
      if ((mchannel[i].stereo) && (!mchannel[i].button_lock))
      {
	bleft += mchannel[i].left;
	bright += mchannel[i].right;
      }
    }
    if (bleft == bright)
      balance = 0;
    else
    {
      if (bleft > bright)
      {
	if (bright == 0)
	  balance = -MAX_MIXER_VALUE;
	else
	  balance = -MAX_MIXER_VALUE+(bright*MAX_MIXER_VALUE/bleft);
      }
      else
      {
	if (bleft == 0)
	  balance = MAX_MIXER_VALUE;
	else
	  balance = MAX_MIXER_VALUE-(bleft*MAX_MIXER_VALUE/bright);
      }
    }
    fl_set_slider_value(overall_balance, balance);

    /* disable, if all channels are locked */
    if (showed_lock_counter == mix.stereo_devs)
    {
      fl_deactivate_object(overall_balance);
      fl_set_object_color(overall_balance, FL_COL1,  INACTIVE_COLOR);
    }
    else
    {
      fl_activate_object(overall_balance);
      fl_set_object_color(overall_balance, FL_COL1,  BALANCE_COLOR);
    }
  }

  /* set lock_all button if necessary */
  if (lock_all != NULL)
  {
    if (button_lock_all != (showed_lock_counter == showed_stereo_counter)) un_lock_all();
    fl_set_button(lock_all, button_lock_all);
  }

  /* set mute_all button if necessary */
  if (mute_all != NULL)
  {
    if (button_mute_all != (showed_silent_counter == show_counter)) un_mute_all();
    fl_set_button(mute_all, button_mute_all);
  }

  /* set global_mute button if necessary */
  if (mute_global != NULL)
  {
    if ( button_global_mute != (global_silent_counter == mix.devs)) un_mute_global();
    fl_set_button(mute_global, button_global_mute);
  }

  /* update all other buttons */
  for (i=0; i<mix.devs; i++)
  {
    if (mchannel[i].fl_lock != NULL) 
      fl_set_button(mchannel[i].fl_lock, mchannel[i].button_lock);
    
    if (mchannel[i].fl_mute != NULL) 
      fl_set_button(mchannel[i].fl_mute, mchannel[i].button_mute);
    
    if (mchannel[i].fl_solo != NULL) 
      fl_set_button(mchannel[i].fl_solo, mchannel[i].button_solo);
  }

  fl_unfreeze_form(mix_form);
  callback_depth--;
}

/* callback for channel mute */
void mix_mute_cb(FL_OBJECT *ob, long data)
{
  callback_depth++;
  fl_freeze_form(mix_form);
  un_mute_channel(data);
  update_mixer_values();
  fl_unfreeze_form(mix_form);
  callback_depth--;
}

/* callback for mute all */
void mix_mute_all_cb(FL_OBJECT *ob, long data) 
{
  callback_depth++;
  fl_freeze_form(mix_form);
  un_mute_all();
  update_mixer_values();
  fl_unfreeze_form(mix_form);
  callback_depth--;
}

/* callback for mute global */
void mix_mute_global_cb(FL_OBJECT *ob, long data) 
{
  callback_depth++;
  fl_freeze_form(mix_form);
  un_mute_global();
  update_mixer_values();
  fl_unfreeze_form(mix_form);
  callback_depth--;
}

/* callback for channel lock */
void mix_lock_cb(FL_OBJECT *ob, long data)
{
  callback_depth++;
  fl_freeze_form(mix_form);
  un_lock_channel(data);
  update_mixer_values();
  fl_unfreeze_form(mix_form);
  callback_depth--;
}

/* callback for lock all */
void mix_lock_all_cb(FL_OBJECT *ob, long data)
{
  callback_depth++;
  fl_freeze_form(mix_form);
  un_lock_all();
  update_mixer_values();
  fl_unfreeze_form(mix_form);
  callback_depth--;
}

/* callback for recording */
void mix_rec_cb(FL_OBJECT *ob, long data)
{
  callback_depth++;
  fl_freeze_form(mix_form);
  un_rec_channel(data);
  update_mixer_values();
  fl_unfreeze_form(mix_form);
  callback_depth--;
}

/* callback for rec all */
void mix_rec_all_cb(FL_OBJECT *ob, long data)
{
  callback_depth++;
  fl_freeze_form(mix_form);
  un_rec_all();
  update_mixer_values();
  fl_unfreeze_form(mix_form);
  callback_depth--;
}

/* callback for channel solo */
void mix_solo_cb(FL_OBJECT *ob, long data)
{
  callback_depth++;
  fl_freeze_form(mix_form);
  un_solo_channel(data);
  update_mixer_values();
  fl_unfreeze_form(mix_form);
  callback_depth--;
}

void mix_slider_cb(FL_OBJECT *ob, long data)       /* callback for mixer sliders */
{
  int vol;

  callback_depth++;
  vol = fl_get_slider_value(ob);

  if ((fl_mouse_button() == 2) || (!mchannel[data].stereo))
  {
    mchannel[data].left = vol;
    mchannel[data].right = vol;

    if (ob == mchannel[data].fl_l)
    {
      if (mchannel[data].fl_r != NULL) 
        fl_set_slider_value(mchannel[data].fl_r, mchannel[data].right);
    }
    else
      fl_set_slider_value(mchannel[data].fl_l, mchannel[data].left);
  }
  else
  {
    if ((fl_mouse_button() == 3) || (mchannel[data].button_lock))
    {
      if (ob == mchannel[data].fl_l)
      {
        if ((mchannel[data].right > mchannel[data].left) &&
	    ((mchannel[data].right + (vol-mchannel[data].left)) > MAX_MIXER_VALUE))
	{
          mchannel[data].left = MAX_MIXER_VALUE-(mchannel[data].right-mchannel[data].left);
          mchannel[data].right = MAX_MIXER_VALUE;
          fl_set_slider_value(mchannel[data].fl_l, mchannel[data].left);
	}
        else
	{
          mchannel[data].right += vol-mchannel[data].left;
          mchannel[data].left = vol;
	}
      }
      else
      {
        if ((mchannel[data].left < mchannel[data].right) &&
	    (mchannel[data].left < (mchannel[data].right-vol)))
	{
          mchannel[data].right = (mchannel[data].right-mchannel[data].left);
          mchannel[data].left = 0;
          fl_set_slider_value(mchannel[data].fl_r, mchannel[data].right);
	}
        else
	{
          mchannel[data].left -= mchannel[data].right-vol;
          mchannel[data].right = vol;
	}
      }
    }
    else
    {
      if (ob == mchannel[data].fl_l)
        mchannel[data].left = vol;
      else
        mchannel[data].right = vol;
    }
  }

  set_volume(data, mchannel[data].left, mchannel[data].right);

  if ((ob != NULL) && !fl_get_button(ob) && 
      (mute_all != NULL) && fl_get_button(mute_all)) fl_set_button(mute_all, 0);

  if ((mchannel[data].left > 0 || mchannel[data].right > 0) && 
      (mchannel[data].button_mute))
  {
    mchannel[data].button_mute = FALSE;
    if (mchannel[data].fl_mute != NULL) fl_set_button(mchannel[data].fl_mute,0);
    if ((mute_all != NULL) && fl_get_button(mute_all)) fl_set_button(mute_all, 0);
  }  
  callback_depth--;
}

void mix_channel_balance_cb(FL_OBJECT *ob, long data)   /* callback for balance slider */
{
  int balance;

  callback_depth++;

  if (fl_mouse_button() == 2)
    balance = 0;
  else
    balance = fl_get_slider_value(ob);

  if (mchannel[data].left > mchannel[data].right)
    mchannel[data].right = mchannel[data].left;
  else
    mchannel[data].left = mchannel[data].right;

  if (balance != 0)
  {
    if (balance < 0)
      mchannel[data].right = (MAX_MIXER_VALUE+balance)*mchannel[data].left/MAX_MIXER_VALUE;
    else
      mchannel[data].left = (MAX_MIXER_VALUE-balance)*mchannel[data].right/MAX_MIXER_VALUE;
  }
    
  fl_set_slider_value(mchannel[data].fl_l, mchannel[data].left);
  fl_set_slider_value(mchannel[data].fl_r, mchannel[data].right);
  set_volume(data, mchannel[data].left, mchannel[data].right);
  callback_depth--;
}

void mix_overall_balance_cb(FL_OBJECT *ob, long data)   /* callback for balance slider */
{
  int i;
  int balance;

  callback_depth++;

  if (fl_mouse_button() == 2)
    balance = 0;
  else
    balance = fl_get_slider_value(ob);

  for (i=0; i<mix.devs; i++)
  {
    if ((mchannel[i].stereo) && (!mchannel[i].button_lock))
    {
      if (mchannel[i].left > mchannel[i].right)
	mchannel[i].right = mchannel[i].left;
      else
	mchannel[i].left = mchannel[i].right;

      if (balance != 0)
      {
	if (balance < 0)
	  mchannel[i].right = (MAX_MIXER_VALUE+balance)*mchannel[i].left/MAX_MIXER_VALUE;
	else
	  mchannel[i].left = (MAX_MIXER_VALUE-balance)*mchannel[i].right/MAX_MIXER_VALUE;
      }
      else
	mchannel[i].left = mchannel[i].right;

      if (mchannel[i].fl_l != NULL)
	fl_set_slider_value(mchannel[i].fl_l, mchannel[i].left);
      if (mchannel[i].fl_r != NULL)
      fl_set_slider_value(mchannel[i].fl_r, mchannel[i].right);

      set_volume(i, mchannel[i].left, mchannel[i].right);
    }
  }
  callback_depth--;
}

/*------------------------------------------------------------------------------------*/

/* app timer */
void sig_alarm_handler(int n)
{
  sig_alarm = TRUE;
}

/* in idle time update meter */
int mix_idle_cb(XEvent *xevent, void *data)
{
  if ((sig_alarm) && (callback_depth == 0))
  {
    update_mixer_values();
    sig_alarm = FALSE;
  }
  return 1;
}

/*------------------------------------------------------------------------------------*/

void mix_end_cb(FL_OBJECT *o, long l)
{
  fl_finish();
  exit(0);
}

void mix_config_cb(FL_OBJECT *o, long l);
void mix_help_cb(FL_OBJECT *o, long l);

void create_form_mix()
{
  #define STEREO_WIDTH   42
  #define MONO_WIDTH     32
  #define SLIDER_HEIGHT  150

  Pixmap icon, icon_shape;
  FL_OBJECT* ob;
  int x, dx, y, i;
  int maxx, maxy;
  int oldx, oldy;
  int showed_channels,
      showed_rec_channels,
      showed_solo_channels,
      showed_stereo_channels,
      showed_balance_channels;
  int org_border_width;
  char *newlabel;

  callback_depth++;

  /* initialize some variables */
  y = oldx = oldy = 0;

  /* if applying new configuration: get old position */
  if (mix_form != NULL)
  {
    oldx = mix_form->x; 
    oldy = mix_form->y;
  }

  /* get original border width */
  org_border_width = fl_get_border_width();

  /* count channels with 'solo' or 'balance' enabled 
     and calculate x and y sizes */
  maxx = 10; /* border */
  showed_channels = 0;
  showed_rec_channels = 0;
  showed_solo_channels = 0;
  showed_stereo_channels = 0;
  showed_balance_channels = 0;
  button_rec_all = TRUE;
  button_mute_all = TRUE;
  button_lock_all = TRUE;
  button_global_mute = TRUE;
  for (i=0; i<mix.devs; i++)
  {
    if (!mchannel[i].button_mute) button_global_mute = FALSE;
    if (mchannel[i].show)
    { 
      showed_channels++;
      if ((mchannel[i].rec_support) && (!mchannel[i].rec_src)) button_rec_all = FALSE;
      if (!mchannel[i].button_mute) button_mute_all = FALSE;
      if (!mchannel[i].button_lock) button_lock_all = FALSE;
      if (mchannel[i].solo_channel) showed_solo_channels++;
      if (mchannel[i].rec_support) showed_rec_channels++;
      if (mchannel[i].stereo) 
      {
	showed_stereo_channels++;
	if (mchannel[i].show_balance) showed_balance_channels++;
	maxx += STEREO_WIDTH;
      }
      else
	maxx += MONO_WIDTH;
    }
  }  
  maxx += 50; /* buttons */

  maxy = 5; /* border */
  maxy += 16; /* label */
  maxy += SLIDER_HEIGHT;
  if (showed_balance_channels > 0) maxy += 25;
  if ((show_overall_balance) && (mix.stereo_devs > 0)) maxy += 25;
  if ((recording_enabled) && (showed_rec_channels > 0)) maxy += 13;
  if ((locking_enabled) && (showed_stereo_channels > 0)) maxy += 13;
  if (muting_enabled) maxy += 13;
  if (showed_solo_channels > 1) maxy += 13;

  /* set all xforms object pointer to NULL */

  overall_balance = NULL;
  rec_all = NULL;
  mute_all = NULL;
  lock_all = NULL;

  for (i=0; i<mix.devs; i++)
  {
    mchannel[i].fl_l = NULL;
    mchannel[i].fl_r = NULL;
    mchannel[i].fl_bal = NULL;
    mchannel[i].fl_rec = NULL;
    mchannel[i].fl_mute = NULL;
    mchannel[i].fl_lock = NULL;
    mchannel[i].fl_solo = NULL;
  }

  /* create form */
  if (mix_form != NULL)
  {
    fl_hide_form(mix_form);
    fl_free_form(mix_form);
  }
  mix_form = fl_bgn_form(FL_FLAT_BOX, maxx, maxy);

  /* add labels and sliders for channels */
  x = 5;
  for (i=0; i<mix.devs; i++)
  {
    if (mchannel[i].show)
    {
      /* x size */
      if (mchannel[i].stereo) dx = STEREO_WIDTH; else dx = MONO_WIDTH;

      /* frame and slider label */
      newlabel = (char *)malloc(strlen(mchannel[i].label));
      strcpy(newlabel, mchannel[i].label);
      while (newlabel[strlen(newlabel)-1]==' ') newlabel[strlen(newlabel)-1]='\0';
      fl_set_border_width(1);

      if (alternate_appearance)
      {
        ob=fl_add_text(FL_NORMAL_TEXT, x-2, 0, dx-3, 16, newlabel);
        fl_set_object_lalign(ob,FL_ALIGN_CENTER);
        fl_set_object_lstyle(ob,FL_BOLD_STYLE);
      }
      else
	ob = fl_add_labelframe(FL_DOWN_FRAME, x, 7, dx-5, SLIDER_HEIGHT+5, newlabel);

      fl_set_border_width(org_border_width);
      free(newlabel);
    
      /* 1. slider */
      if (alternate_appearance)
        mchannel[i].fl_l=ob=fl_add_valslider(FL_VERT_BROWSER_SLIDER, x, 16,
                                             (mchannel[i].stereo) ? (dx-3)/2 : dx-5, SLIDER_HEIGHT, "");
      else
      {
        mchannel[i].fl_l=ob=fl_add_valslider(FL_VERT_NICE_SLIDER, x, 12,
					     (mchannel[i].stereo) ? (dx-5)/2 : dx-5, SLIDER_HEIGHT, "");
	fl_set_object_boxtype(ob, FL_FLAT_BOX);
      } 
      fl_set_slider_return(ob, FL_RETURN_CHANGED);
      fl_set_object_callback(ob, mix_slider_cb, i);
      fl_set_slider_bounds(ob, MAX_MIXER_VALUE, 0);
      fl_set_slider_increment(ob, 1, 1);
      fl_set_slider_precision(ob, 0);
      fl_set_slider_step(ob, 1);
        
      /* 2. slider */
      if (mchannel[i].stereo)
      {
	if (alternate_appearance)
	  mchannel[i].fl_r=ob=fl_add_valslider(FL_VERT_BROWSER_SLIDER,x+(dx-3)/2,16,(dx-3)/2,SLIDER_HEIGHT,"");
	else
        {
	  mchannel[i].fl_r=ob=fl_add_valslider(FL_VERT_NICE_SLIDER,x+(dx-5)/2,12,(dx-5)/2,SLIDER_HEIGHT,"");
	  fl_set_object_boxtype(ob, FL_FLAT_BOX);
	}
	fl_set_slider_return(ob, FL_RETURN_CHANGED);
	fl_set_object_callback(ob, mix_slider_cb, i);
	fl_set_slider_bounds(ob, MAX_MIXER_VALUE, 0);
	fl_set_slider_increment(ob, 1, 1);
	fl_set_slider_precision(ob, 0);
	fl_set_slider_step(ob, 1);
      }

      y = SLIDER_HEIGHT + 16;

      /* channel balance slider */
      if ((mchannel[i].stereo) && (mchannel[i].show_balance))
      {
	if (alternate_appearance)
	  mchannel[i].fl_bal=ob=fl_add_slider(FL_HOR_BROWSER_SLIDER,x,y,dx-5,18,"");
	else
	{
	  fl_set_border_width(1);
	  ob = fl_add_frame(FL_DOWN_FRAME, x, y, dx-5, 20, "");
	  fl_set_border_width(org_border_width);
	  mchannel[i].fl_bal=ob=fl_add_slider(FL_HOR_NICE_SLIDER,x,y+1,dx-5,18,"");
	  fl_set_object_boxtype(ob, FL_FLAT_BOX);
	}
	fl_set_slider_return(ob, FL_RETURN_CHANGED);
	fl_set_object_callback(ob, mix_channel_balance_cb, i);
	fl_set_slider_bounds(ob, -MAX_MIXER_VALUE, MAX_MIXER_VALUE);
	fl_set_object_color(ob, FL_COL1, BALANCE_COLOR);
	fl_set_slider_increment(ob, 1, 1);
	fl_set_slider_precision(ob, 0);
	fl_set_slider_step(ob, 1);
      }	

      /* calc new x pos */
      x += dx;
    }
  }
  if (showed_balance_channels > 0) y += 25;

  /* add overall balance slider */
  if ((show_overall_balance) && (mix.stereo_devs > 0))
  {
    if (alternate_appearance)
      overall_balance=ob=fl_add_slider(FL_HOR_BROWSER_SLIDER,5,y,x-10,18,"");
    else
    {
      fl_set_border_width(1);
      ob = fl_add_frame(FL_DOWN_FRAME, 5, y, x-10, 20, "");
      fl_set_border_width(org_border_width);
      overall_balance=ob=fl_add_slider(FL_HOR_NICE_SLIDER,5,y+1,x-10,18,"");
      fl_set_object_boxtype(ob, FL_FLAT_BOX);
    }
    fl_set_slider_return(ob, FL_RETURN_CHANGED);
    fl_set_object_callback(ob, mix_overall_balance_cb, 0);
    fl_set_slider_bounds(ob, -MAX_MIXER_VALUE, MAX_MIXER_VALUE);
    fl_set_object_color(ob, FL_COL1, BALANCE_COLOR);
    fl_set_slider_increment(ob, 1, 1);
    fl_set_slider_precision(ob, 0);
    fl_set_slider_step(ob, 1);

    y += 25;
  }

  /* add rec, lock, mute and solo buttons */

  /* rec buttons */
  if (recording_enabled)
  {
    x = 5;
    fl_bgn_group();
    for (i=0; i<mix.devs; i++)
    {
      if (mchannel[i].show)
      {
	/* x size */
	if (mchannel[i].stereo) dx = STEREO_WIDTH; else dx = MONO_WIDTH;

	/* button */
	if (mchannel[i].rec_support)
	{
          if (exclusive_rec_selection) 
	    mchannel[i].fl_rec=ob=fl_add_mybutton(FL_RADIO_BUTTON, x, y, dx, 12, "rec");
	  else
	    mchannel[i].fl_rec=ob=fl_add_mybutton(FL_PUSH_BUTTON, x, y, dx, 12, "rec");
	  fl_set_object_lcolor(ob, REC_COLOR);
	  fl_set_object_color(ob, FL_COL1, REC_COLOR);
	  fl_set_object_callback(ob, mix_rec_cb, i);
	  fl_set_object_lsize(ob, FL_AUTO_SIZE);
	}
	/* calc new x pos */
	x += dx;
      }
    }
    if (showed_rec_channels > 0) y += 13;
    fl_end_group();
  }

  /* lock buttons */
  if (locking_enabled)
  {
    x = 5;
    fl_bgn_group();
    for (i=0; i<mix.devs; i++)
    {
      if (mchannel[i].show)
      {
	/* x size */
	if (mchannel[i].stereo) dx = STEREO_WIDTH; else dx = MONO_WIDTH;

	/* button */
	if (mchannel[i].stereo)
        {
	  mchannel[i].fl_lock=ob=fl_add_mybutton(FL_PUSH_BUTTON, x, y, dx, 12, "lock");
	  if (mchannel[i].button_lock) fl_set_button(ob, 1);
	  fl_set_object_callback(ob, mix_lock_cb, i);
	  fl_set_object_lcolor(ob, LOCK_COLOR);
	  fl_set_object_color(ob, FL_COL1, LOCK_COLOR); 
          fl_set_object_lsize(ob, FL_AUTO_SIZE);
	}
	/* calc new x pos */
	x += dx;
      }
    }
    if (showed_stereo_channels > 0) y += 13;
    fl_end_group();
  }

  /* mute buttons */
  if (muting_enabled)
  {
    x = 5;
    fl_bgn_group();
    for (i=0; i<mix.devs; i++)
    {
      if (mchannel[i].show)
      {
	/* x size */
	if (mchannel[i].stereo) dx = STEREO_WIDTH; else dx = MONO_WIDTH;

	/* button */
	mchannel[i].fl_mute=ob=fl_add_mybutton(FL_PUSH_BUTTON, x, y, dx, 12, "mute");
	if (mchannel[i].button_mute) fl_set_button(ob, 1);
	fl_set_object_lcolor(ob, MUTE_COLOR);
	fl_set_object_color(ob, FL_COL1, MUTE_COLOR);
	fl_set_object_callback(ob, mix_mute_cb, i);
        fl_set_object_lsize(ob, FL_AUTO_SIZE);

	/* calc new x pos */
	x += dx;
      }
    }
    y += 13;  
    fl_end_group();
  }

  /* solo buttons */
  if (showed_solo_channels > 1)
  {
    x = 5;
    fl_bgn_group();
    for (i=0; i<mix.devs; i++)
    {
      if (mchannel[i].show)
      {
	/* x size */
	if (mchannel[i].stereo) dx = STEREO_WIDTH; else dx = MONO_WIDTH;

	/* button */
	if (mchannel[i].solo_channel)
	{
	  mchannel[i].fl_solo=ob=fl_add_mybutton(FL_RADIO_BUTTON, x, y, dx, 12, "solo");
	  if (mchannel[i].button_solo) fl_set_button(ob, 1);
          fl_set_object_lcolor(ob, SOLO_COLOR);
	  fl_set_object_color(ob, FL_COL1, SOLO_COLOR);
	  fl_set_object_callback(ob, mix_solo_cb, i);
	  fl_set_object_lsize(ob, FL_AUTO_SIZE);
	}

	/* calc new x pos */
	x += dx;
      }
    }
    y += 13;
    fl_end_group();
  }

  /* help, config, quit buttons */
  help_button=ob=fl_add_button(FL_NORMAL_BUTTON, x, 5, 50, 20, "Help");
    fl_set_object_callback(ob, mix_help_cb, 0);
    fl_set_object_shortcut(ob, "Hh?", TRUE);
    fl_set_object_gravity(ob,FL_NorthEast, FL_NorthEast);
    if (help_form != NULL)
    {  
      fl_deactivate_object(help_button);
      fl_set_object_lcol(help_button, FL_INACTIVE);
    }
  config_button=ob=fl_add_button(FL_NORMAL_BUTTON, x, 30, 50, 20, "Config");
    fl_set_object_callback(ob, mix_config_cb, 0);
    fl_set_object_shortcut(ob, "Cc", TRUE);
    fl_set_object_gravity(ob,FL_NorthEast, FL_NorthEast);
    if (config_form != NULL)
    {  
      fl_deactivate_object(config_button);
      fl_set_object_lcol(config_button, FL_INACTIVE);
    }
  ob=fl_add_button(FL_NORMAL_BUTTON, x, 55, 50, 20, "Quit");
    fl_set_object_callback(ob, mix_end_cb, 0);
    fl_set_object_shortcut(ob, "Qq#q#x", TRUE);
    fl_set_object_gravity(ob,FL_NorthEast, FL_NorthEast);

  /* the 'global' button(s) */
  mute_global=ob=fl_add_lightbutton(FL_PUSH_BUTTON, x, 90, 50, 20, "Mute");
    fl_set_object_callback(ob, mix_mute_global_cb, 0);
    fl_set_object_gravity(ob,FL_NorthEast, FL_NorthEast);
    fl_set_object_color(ob, FL_COL1, MUTE_COLOR);
    fl_set_object_shortcut(ob, "Mm#m", TRUE);
    fl_set_button(ob, button_global_mute);

  /* the 'all' buttons */
  y = SLIDER_HEIGHT + 16;
  if (showed_balance_channels > 0) y += 25;
  if ((show_overall_balance) && (mix.stereo_devs > 0)) y += 25;

  if ((recording_enabled) && (showed_rec_channels > 1))
  {
    if (!exclusive_rec_selection)
    {
      rec_all=ob=fl_add_mybutton(FL_PUSH_BUTTON, x, y, 50, 12, "rec all");
      fl_set_object_callback(ob, mix_rec_all_cb, 0);
      fl_set_object_lcolor(ob, REC_COLOR);
      fl_set_object_color(ob, FL_COL1, REC_COLOR);
      fl_set_object_lsize(ob, FL_AUTO_SIZE);
      fl_set_object_shortcut(ob, "rR#r", 0);
      fl_set_button(ob, button_rec_all);
    }
    y += 13;
  }
  else
    rec_all = NULL;

  if ((locking_enabled) && (showed_stereo_channels > 1))
  {
    lock_all=ob=fl_add_mybutton(FL_PUSH_BUTTON, x, y, 45, 12, "lock all");
    fl_set_object_callback(ob, mix_lock_all_cb, 0);
    fl_set_object_lcolor(ob, LOCK_COLOR);
    fl_set_object_color(ob, FL_COL1, LOCK_COLOR); 
    fl_set_object_lsize(ob, FL_AUTO_SIZE);
    fl_set_object_shortcut(ob, "lL#l", 0);
    fl_set_button(ob, button_lock_all);
    y += 13;
  }
  else
    lock_all = NULL;

  if (muting_enabled) 
  {
    if (showed_channels > 1)
    {
      mute_all=ob=fl_add_mybutton(FL_PUSH_BUTTON, x, y, 45, 12, "mute all");
      fl_set_object_callback(ob, mix_mute_all_cb, 0);
      fl_set_object_lcolor(ob, MUTE_COLOR);
      fl_set_object_color(ob, FL_COL1, MUTE_COLOR);
      fl_set_object_lsize(ob, FL_AUTO_SIZE);
      fl_set_object_shortcut(ob, "uU#u", 0);
      fl_set_button(ob, button_mute_all);
    }
    y += 13;
  }
  else
    mute_all = NULL;

  fl_end_form();
  fl_set_form_minsize(mix_form, maxx, maxy);

  /* force update */
  update_mixer_values();

  /* now show the form (with nice title)*/
  newlabel = (char *)malloc(strlen(mix.info.name)+strlen(PRG_NAME)+10);
  sprintf(newlabel, "%s for %s", PRG_NAME, mix.info.name);
  if (config_form != NULL)
  {
    fl_set_form_position(mix_form, oldx, oldy);
    fl_show_form(mix_form, FL_PLACE_GEOMETRY, FL_FULLBORDER, newlabel);
  }
  else
    fl_show_form(mix_form, FL_PLACE_FREE, FL_FULLBORDER, newlabel);
  free(newlabel);

  /* set my icon */
  XpmCreatePixmapFromData( fl_get_display(), mix_form->window, smix_xpm,
                           &icon, &icon_shape, NULL);
  fl_winicon(mix_form->window, icon, icon_shape);

  callback_depth--;

  /* if active update: set timer */
  if (active_update)
    fl_set_idle_callback(mix_idle_cb, NULL);
  else
    fl_set_idle_callback(NULL, NULL);
}

/*------------------------------------------------------------------------------------*/

void mix_help_close_cb(FL_OBJECT *ob, long l)
{
  fl_set_object_lcol(help_button, FL_BLACK);
  fl_activate_object(help_button);
  fl_hide_form(help_form);
  fl_free_form(help_form);
  help_form = NULL;
}

void mix_help_cb(FL_OBJECT *ob, long l)
{
  FL_OBJECT *obj;

  fl_deactivate_object(help_button);
  fl_set_object_lcol(help_button, FL_INACTIVE);

  help_form = fl_bgn_form(FL_UP_BOX, 400, 260);

  fl_add_box(FL_FLAT_BOX,0,0,400,260,"");

  obj = fl_add_text(FL_NORMAL_TEXT,5,0,390,30,PRG_NAME " " VERSION);
    fl_set_object_lcol(obj,FL_BLUE);
    fl_set_object_lsize(obj,FL_HUGE_SIZE);
    fl_set_object_align(obj,FL_ALIGN_TOP_LEFT);
    fl_set_object_lstyle(obj,FL_TIMESBOLDITALIC_STYLE+FL_EMBOSSED_STYLE);
    fl_set_object_gravity(obj,FL_NorthWest, FL_NorthEast);

  obj = fl_add_text(FL_NORMAL_TEXT,5,30,390,60,
		    "xforms based mixer for Linux/X11R6\n"
		    "(C)opyright 1997 by Sven Anders (anderss@fmi.uni-passau.de)");
    fl_set_object_lcol(obj,FL_BLUE);
    fl_set_object_align(obj,FL_ALIGN_TOP_LEFT);
    fl_set_object_lstyle(obj,FL_TIMESBOLDITALIC_STYLE);

  fl_add_frame(FL_DOWN_FRAME,5,70,388,160,"");

  obj = fl_add_text(FL_NORMAL_TEXT,10,75,90,20,"L");
    fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,110,75,280,20,"(un) lock all channels");
    fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);

  obj = fl_add_text(FL_NORMAL_TEXT,10,95,90,20,"M");
    fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,110,95,280,20,"global (un) mute on all channels");
    fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);

  obj = fl_add_text(FL_NORMAL_TEXT,10,115,90,20,"U");
    fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,110,115,280,20,"(un) mute on all shown channels");
    fl_set_object_lalign(obj,FL_ALIGN_LEFT|FL_ALIGN_INSIDE);

  obj = fl_add_text(FL_NORMAL_TEXT,10,135,90,20,"R");
    fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,110,135,280,40,"(de) activate all recording sources\n(when recording sources aren't exclusive)");
    fl_set_object_lalign(obj,FL_ALIGN_TOP_LEFT|FL_ALIGN_INSIDE);

  obj = fl_add_text(FL_NORMAL_TEXT,10,165,90,20,"Mouse button 1");
    fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,110,165,280,20,"change single channel  or  change balance");
    fl_set_object_lalign(obj,FL_ALIGN_TOP_LEFT|FL_ALIGN_INSIDE);

  obj = fl_add_text(FL_NORMAL_TEXT,10,185,90,20,"Mouse button 2");
    fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,110,185,280,20,"change both channels and center the balance");
    fl_set_object_lalign(obj,FL_ALIGN_TOP_LEFT|FL_ALIGN_INSIDE);

  obj = fl_add_text(FL_NORMAL_TEXT,10,205,90,20,"Mouse button 3");
    fl_set_object_lalign(obj,FL_ALIGN_RIGHT|FL_ALIGN_INSIDE);
    fl_set_object_lstyle(obj,FL_BOLD_STYLE);
  obj = fl_add_text(FL_NORMAL_TEXT,110,205,280,20,"change both channels without centering the balance");
    fl_set_object_lalign(obj,FL_ALIGN_TOP_LEFT|FL_ALIGN_INSIDE);

  obj=fl_add_button(FL_NORMAL_BUTTON,330,235,65,20,"Close");
    fl_set_object_callback(obj, mix_help_close_cb, 0);
    fl_set_object_shortcut(obj, "#c", TRUE);

  fl_end_form();
  fl_show_form(help_form, FL_PLACE_MOUSE, FL_TRANSIENT, PRG_NAME " help/about");
}

/*------------------------------------------------------------------------------------*/

struct config_channel
{
  FL_OBJECT *fl_label;
  FL_OBJECT *fl_show;
  FL_OBJECT *fl_balance;
  FL_OBJECT *fl_solo;
} cchannel[MAX_NUM_MIXER_DEVS];

FL_OBJECT *show_mute_button,
          *show_lock_button,
          *show_overall_button,
          *show_rec_button,
          *exclusive_rec_button,
          *alternate_app_button,
          *active_update_button,
          *save_volumes_button,
          *save_recsrc_button,
          *save_channel_button,
          *save_app_button;

#define KEY_COMMENT       "#"
#define KEY_DEVICE        "device"
#define KEY_MIXERID       "mixerid"
#define KEY_ACTIVE_UPDATE "active update"
#define KEY_CHANNEL       "channel"
#define KEY_VOLUME        "volume"
#define KEY_RECSRC        "recsrc"
#define KEY_SHOW_MUTE     "show mute"
#define KEY_SHOW_LOCK     "show lock"
#define KEY_SHOW_REC      "show rec"
#define KEY_EXCL_REC      "exclusive recs"
#define KEY_SHOW_BALANCE  "show overall balance"
#define KEY_ALT_APPEAR    "alternate appearance"

void check_config_filename()
{
  struct passwd *users_entry;
  char *newname;

  if ((strlen(config_file) > 1) && (config_file[0] == '~'))
  {
    users_entry = getpwuid(getuid());
    newname = (char *)malloc(strlen(users_entry->pw_dir)+strlen(config_file)+2);
    strcpy(newname, users_entry->pw_dir);
    strcat(newname, config_file+1);
    config_file = newname;
  }
}

void mix_config_load(void)
{
  long length, read;
  char *begin, *end, *end2;
  char *data;
  FILE *rfile;
  int dev_num;
  int channel = 0;
  int i, found;

  /* read file */

  check_config_filename();
  if (verbose) printf("loading configuration file '%s' ...\n", config_file);

  rfile = fopen(config_file, "r");
  if (rfile != NULL)
  {
    /* determine length of file */
    fseek(rfile, 0L, SEEK_END);
    length = ftell(rfile);
    fseek(rfile, 0L, SEEK_SET);
    
    /* allocate memory for data */
    data = (char *)malloc(length+2);
    if (data != NULL)
    {
      data[0] = 0;
      read = fread(data, length, sizeof(char), rfile);
      if (read == 0)
      {
        free(data); data = NULL; length = 0;
      }
      else
      {
       data[length] = '\n'; data[length+1] = 0;
      }   
    }
    fclose(rfile);
  }  

  if ((rfile == NULL) || (data == NULL))
  {
    if (errno == ENOENT)
      {if (verbose) printf("configuration file '%s' not found (using defaults).\n", config_file);}
    else
      fprintf(stderr, "error (%i) reading configuration file '%s'. (using defaults)\n", errno, config_file);
    return;
  }

  /* check for correct mixer ID */

  found = FALSE;
  begin = data; end = data;
  while ((found == FALSE) && (end != data+length))
  {
    end = strchr(begin, '\n');
    if ((begin[0] != '\0') && 
	(strncmp(begin, KEY_COMMENT, strlen(KEY_COMMENT)) != 0))
    {
      if (strncmp(begin, KEY_MIXERID, strlen(KEY_MIXERID)) == 0)
      {
	found = TRUE;
	begin += (strlen(KEY_MIXERID)+1);
	if (strncmp(begin, mix.info.id, max(end-begin, strlen(mix.info.id))) != 0)
	{
	  fprintf(stderr, "Mixer IDs does not match. Ignoring configuration settings.\n");
	  free(data);
	  return;
	}
      }
    }
    begin = end+1;
  }

  if (!found)
  {
    fprintf(stderr, "No mixer ID found in file. Ignoring configuration settings.\n");
    free(data);
    return;
  }

  /* set all configurable parameters to FALSE */

  muting_enabled = FALSE;
  locking_enabled = FALSE;
  recording_enabled = FALSE;
  exclusive_rec_selection = FALSE;
  show_overall_balance = FALSE;
  alternate_appearance = FALSE;
  active_update = FALSE;

  for (i=0; i<mix.devs; i++)
    if (mchannel[i].rec_src) un_rec_channel(i);

  /* parse file */
  begin = data; end = data;

  while (end != data+length)
  {
    end = strchr(begin, '\n');
    end[0]=0;

    if ( (begin[0] != '\0') && (strncmp(begin, KEY_COMMENT, strlen(KEY_COMMENT)) != 0))
    {
      /* mixer device */
      if (strncmp(begin, KEY_DEVICE, strlen(KEY_DEVICE)) == 0)
      {
	begin += strlen(KEY_DEVICE);
	while ((*begin == ' ') && (*begin != '\n') && (*begin != '\0')) begin++;
	if ((*begin != '\n') && (*begin != '\0'))
	{
	  dev_name = (char *)malloc(strlen(begin));
	  strcpy(dev_name, begin);
	  init_channel_struct();
	}
      }

      /* muting */
      if (strncmp(begin, KEY_SHOW_MUTE, strlen(KEY_SHOW_MUTE)) == 0)
        muting_enabled = TRUE;

      /* locking */
      if (strncmp(begin, KEY_SHOW_LOCK, strlen(KEY_SHOW_LOCK)) == 0)
	locking_enabled = TRUE;

      /* recording */
      if (strncmp(begin, KEY_SHOW_REC, strlen(KEY_SHOW_REC)) == 0)
	recording_enabled = TRUE;

      /* exclusive recording sources */
      if (strncmp(begin, KEY_EXCL_REC, strlen(KEY_EXCL_REC)) == 0)
	exclusive_rec_selection = TRUE;

      /* show overall balance */
      if (strncmp(begin, KEY_SHOW_BALANCE, strlen(KEY_SHOW_BALANCE)) == 0)
	show_overall_balance = TRUE;

      /* alternate appearance */
      if (strncmp(begin, KEY_ALT_APPEAR, strlen(KEY_ALT_APPEAR)) == 0)
	alternate_appearance = TRUE;

      /* alternate appearance */
      if (strncmp(begin, KEY_ACTIVE_UPDATE, strlen(KEY_ACTIVE_UPDATE)) == 0)
	active_update = TRUE;

      /* channel settings */
      if (strncmp(begin, KEY_CHANNEL, strlen(KEY_CHANNEL)) == 0)
      {
	begin += strlen(KEY_CHANNEL);
	while ((*begin == ' ') && (*begin != '\n')) begin++;
	if (*begin != '\n')
	{
	  /*device number*/
	  end2 = strchr(begin, ' '); end2[0]=0;
	  dev_num = atoi(begin); begin = end2+1;
	  found = FALSE;
	  for (i=0; i<mix.devs; i++)
	    if (mchannel[i].dev_num == dev_num)
            {
	      found = TRUE; channel = i;
	    }
	  if (found)
	  {
	    /*show*/
	    end2 = strchr(begin, ' '); end2[0]=0;
	    mchannel[channel].show = atoi(begin);
	    begin = end2+1;
	    /*balance*/
	    end2 = strchr(begin, ' '); end2[0]=0;
	    mchannel[channel].show_balance = atoi(begin);
	    begin = end2+1;
	    /*solo*/
	    end2 = strchr(begin, ' '); 
	    if (end2 != NULL) end2[0]=0;
	    mchannel[channel].solo_channel = atoi(begin);
	    if (end2 != NULL)
	    {
	      /*label (optional)*/
	      begin = end2+1;
	      if (mchannel[channel].label != NULL) free(mchannel[channel].label);
	      mchannel[channel].label = (char *)malloc(strlen(begin)+1);
	      strcpy(mchannel[channel].label, begin);
	    }
	  }
        }
      }

      /* volume settings */
      if (strncmp(begin, KEY_VOLUME, strlen(KEY_VOLUME)) == 0)
      {
	begin += strlen(KEY_VOLUME);
	while ((*begin == ' ') && (*begin != '\n')) begin++;
	if (*begin != '\n')
	{
	  /*device number*/
	  end2 = strchr(begin, ' '); end2[0]=0;
	  dev_num = atoi(begin); begin = end2+1;
	  found = FALSE;
	  for (i=0; i<mix.devs; i++)
	    if (mchannel[i].dev_num == dev_num)
            {
	      found = TRUE; channel = i;
	    }
	  if (found)
	  {
	    /*left/mono*/
	    end2 = strchr(begin, ' ');
	    if (end2 == NULL)
	    {
	      mchannel[channel].left = atoi(begin);
	      mchannel[channel].right = atoi(begin);
	    }
	    else
	    {
	      end2[0]=0;
	      mchannel[channel].left = atoi(begin);
	      begin = end2+1;
	      /*right*/
	      mchannel[channel].right = atoi(begin);
	    }
	    set_volume(channel,mchannel[channel].left,mchannel[channel].right);
	  }
        }
      }

      /* recsrc settings */
      if (strncmp(begin, KEY_RECSRC, strlen(KEY_RECSRC)) == 0)
      {
	begin += strlen(KEY_RECSRC);
	while ((*begin == ' ') && (*begin != '\n')) begin++;
	dev_num = atoi(begin);
	for (i=0; i<mix.devs; i++)
	  if (mchannel[i].dev_num == dev_num) un_rec_channel(i);
      }

    } /* if no comment */
       
    begin=end+1;

  } /*while*/

  free(data);
}

void mix_config_save_cb(FL_OBJECT *ob, long l)
{
  int i;
  FILE *sfile;

  check_config_filename();
  if ( (sfile = fopen(config_file, "wb+")) != NULL )
  {
    fprintf(sfile, KEY_COMMENT "-------------------------------------------------------------------------------\n");
    fprintf(sfile, KEY_COMMENT " " PRG_NAME " Version " VERSION " - Configuration File (for %s)\n", mix.info.name );
    fprintf(sfile, KEY_COMMENT "-------------------------------------------------------------------------------\n");
    fprintf(sfile, KEY_MIXERID " %s\n", mix.info.id);

    fprintf(sfile, "\n" KEY_COMMENT "-------------------------------------------------------------------------------\n");
    fprintf(sfile,  KEY_COMMENT " default device and update stettings\n");
    fprintf(sfile,  "device %s\n", dev_name);
    if (!fl_get_button(active_update_button)) fprintf(sfile, KEY_COMMENT);
      fprintf(sfile, KEY_ACTIVE_UPDATE "\n");

    if (fl_get_button(save_app_button))
    {
      fprintf(sfile, "\n" KEY_COMMENT "-------------------------------------------------------------------------------\n");
      fprintf(sfile,  KEY_COMMENT " appearance settings\n\n");

      if (!fl_get_button(show_mute_button)) fprintf(sfile, KEY_COMMENT);
        fprintf(sfile, KEY_SHOW_MUTE "\n");
      if (!fl_get_button(show_lock_button)) fprintf(sfile, KEY_COMMENT);
        fprintf(sfile, KEY_SHOW_LOCK "\n");
      if (!fl_get_button(show_rec_button)) fprintf(sfile, KEY_COMMENT);
        fprintf(sfile, KEY_SHOW_REC "\n");
      if (!fl_get_button(exclusive_rec_button)) fprintf(sfile, KEY_COMMENT);
        fprintf(sfile, KEY_EXCL_REC "\n");
      if (!fl_get_button(show_overall_button)) fprintf(sfile, KEY_COMMENT);
        fprintf(sfile, KEY_SHOW_BALANCE "\n");
      if (!fl_get_button(alternate_app_button)) fprintf(sfile, KEY_COMMENT);
        fprintf(sfile, KEY_ALT_APPEAR "\n");
    }

    fprintf(sfile, "\n" KEY_COMMENT "-------------------------------------------------------------------------------\n");
    fprintf(sfile, KEY_COMMENT " channel / volume / recording source settings\n");
    fprintf(sfile, KEY_COMMENT " channel <device number> <show> <balance> <solo> [<label text>]\n");
    fprintf(sfile, KEY_COMMENT " volume <device number> <left | mono> [<right>]\n");
    fprintf(sfile, KEY_COMMENT " recsrc <device number>\n");

    if (fl_get_button(save_channel_button))
    {
      fprintf(sfile, "\n");
      for (i=0; i<mix.devs; i++)
      {
	fprintf(sfile, KEY_CHANNEL " %u", mchannel[i].dev_num);
	fprintf(sfile, " %u", fl_get_button(cchannel[i].fl_show));
	if (cchannel[i].fl_balance != NULL)
	  fprintf(sfile, " %u", fl_get_button(cchannel[i].fl_balance));
        else
	  fprintf(sfile, " 0");
	fprintf(sfile, " %u", fl_get_button(cchannel[i].fl_solo));
	fprintf(sfile, " %s", fl_get_input(cchannel[i].fl_label));
	fprintf(sfile, "\n");
      }
    }

    if (fl_get_button(save_volumes_button))	
    {
      fprintf(sfile, "\n");
      for (i=0; i<mix.devs; i++)
      {
	fprintf(sfile, KEY_VOLUME " %u %u", mchannel[i].dev_num, mchannel[i].left);
	if (mchannel[i].stereo)
	  fprintf(sfile, " %u\n", mchannel[i].right); 
	else
	  fprintf(sfile, "\n");	  
      }
    }

    if (fl_get_button(save_recsrc_button))
    {
      fprintf(sfile, "\n");
      for (i=0; i<mix.devs; i++)
      {
	if (mchannel[i].rec_src) fprintf(sfile, KEY_RECSRC " %u\n", mchannel[i].dev_num);
      }
    }

    fprintf(sfile, "\n" KEY_COMMENT "-------------------------------------------------------------------------------\n\n");

    fclose(sfile);
  }

  if (errno)
    fl_show_oneliner( "Error writing settings!", 
		      ob->form->x+ob->x, ob->form->y+ob->y-25 );
  else
    fl_show_oneliner( "Settings successfully saved.", 
		      ob->form->x+ob->x, ob->form->y+ob->y-25 );
  XFlush(fl_get_display());
  my_msleep(1000);
  fl_hide_oneliner();
}

void mix_config_apply_cb(FL_OBJECT *ob, long l)
{
  int i;

  for (i=0; i<mix.devs; i++)
  {
    if (mchannel[i].label != NULL) free(mchannel[i].label);
    mchannel[i].label = (char *)malloc(strlen(fl_get_input(cchannel[i].fl_label))+1);
    strcpy(mchannel[i].label, fl_get_input(cchannel[i].fl_label));
    mchannel[i].show = fl_get_button(cchannel[i].fl_show);
    if (cchannel[i].fl_balance != NULL)
      mchannel[i].show_balance = fl_get_button(cchannel[i].fl_balance);
    mchannel[i].solo_channel = fl_get_button(cchannel[i].fl_solo);
  }

  muting_enabled = fl_get_button(show_mute_button);
  locking_enabled = fl_get_button(show_lock_button);
  recording_enabled = fl_get_button(show_rec_button);
  exclusive_rec_selection = fl_get_button(exclusive_rec_button);
  show_overall_balance = fl_get_button(show_overall_button);
  alternate_appearance = fl_get_button(alternate_app_button);
  active_update = fl_get_button(active_update_button);

  create_form_mix();
}

void mix_config_dummy_cb(FL_OBJECT *obj, long l)
{ }

void mix_config_close_cb(FL_OBJECT *obj, long l)
{
  mix_config_apply_cb(NULL, 0);
  fl_set_object_lcol(config_button, FL_BLACK);
  fl_activate_object(config_button);
  fl_hide_form(config_form);
  fl_free_form(config_form);
  config_form = NULL;
}

void mix_config_channel_show_cb(FL_OBJECT *obj, long i)
{
  fl_freeze_form(config_form);

  if (fl_get_button(obj))
  {
    if (cchannel[i].fl_balance != NULL)
    {
      fl_activate_object(cchannel[i].fl_balance);
      fl_set_object_color(cchannel[i].fl_balance, FL_COL1,  BALANCE_COLOR); 
    }
    fl_activate_object(cchannel[i].fl_solo);
    fl_set_object_color(cchannel[i].fl_solo, FL_COL1,  SOLO_COLOR); 
  }
  else
  {
    if (cchannel[i].fl_balance != NULL)
    {
      fl_deactivate_object(cchannel[i].fl_balance);
      fl_set_object_color(cchannel[i].fl_balance, FL_COL1,  INACTIVE_COLOR);
    }
    fl_deactivate_object(cchannel[i].fl_solo);
    fl_set_object_color(cchannel[i].fl_solo, FL_COL1,  INACTIVE_COLOR);
  }
  fl_unfreeze_form(config_form);
}


void mix_config_cb(FL_OBJECT *ob, long l)
{
  FL_OBJECT *obj;
  char int_string[42];
  int addy = 0;
  int i;

  fl_deactivate_object(config_button);
  fl_set_object_lcol(config_button, FL_INACTIVE);

  if (mix.devs>17) addy = (mix.devs-17)*20;
  config_form = fl_bgn_form(FL_FLAT_BOX, 500, 405+addy);
  obj = fl_add_box(FL_FLAT_BOX,0,0,500,405+addy,"");

  obj = fl_add_labelframe(FL_ENGRAVED_FRAME,10,10,230,385+addy,"Channels");
  obj = fl_add_text(FL_NORMAL_TEXT,90,20,40,20,"show");
    fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
  obj = fl_add_text(FL_NORMAL_TEXT,135,20,50,20,"balance");
    fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
  obj = fl_add_text(FL_NORMAL_TEXT,195,20,30,20,"solo");
    fl_set_object_lalign(obj,FL_ALIGN_CENTER|FL_ALIGN_INSIDE);

  for (i=0; i<mix.devs; i++)
  {
    sprintf(int_string, "%u:", i);
    cchannel[i].fl_label=obj=fl_add_input(FL_NORMAL_INPUT,35,i*20+40,55,20,int_string);
      fl_set_input(obj, mchannel[i].label);
      fl_set_object_callback(obj, mix_config_dummy_cb, 0);
      fl_set_object_boxtype(obj,FL_FLAT_BOX);

    cchannel[i].fl_show=obj=fl_add_checkbutton(FL_PUSH_BUTTON,100,i*20+40,22,22,"");
      fl_set_object_callback(obj, mix_config_channel_show_cb, i);
      if (mchannel[i].show) fl_set_button(obj,1);

    if (mchannel[i].stereo)
    {
      cchannel[i].fl_balance=obj=fl_add_checkbutton(FL_PUSH_BUTTON,150,i*20+40,22,22,"");
        fl_set_object_callback(obj, mix_config_dummy_cb, 0);
        if (mchannel[i].show_balance) fl_set_button(obj,1);
    }
    else
      cchannel[i].fl_balance=NULL;

    cchannel[i].fl_solo=obj=fl_add_checkbutton(FL_PUSH_BUTTON,200,i*20+40,22,22,"");
      fl_set_object_callback(obj, mix_config_dummy_cb, 0);
      if (mchannel[i].solo_channel) fl_set_button(obj,1);

    /* activate/deactivate balance/show buttons */
    mix_config_channel_show_cb(cchannel[i].fl_show, i);
  }


  obj = fl_add_labelframe(FL_ENGRAVED_FRAME,250,10,240,160,"Appearence");
  show_mute_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,255,20,220,22,"show channel mute buttons");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
    if (muting_enabled) fl_set_button(obj,1);
  show_lock_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,255,45,220,22,"show channel lock buttons");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
    if (locking_enabled) fl_set_button(obj,1);
  show_overall_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,255,115,220,22,"show overall balance slider");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
    if (show_overall_balance) fl_set_button(obj,1);
  show_rec_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,255,70,220,22,"show record source selection buttons");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
    if (recording_enabled) fl_set_button(obj,1);
  exclusive_rec_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,275,90,210,22,"recording source buttons are exclusive");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
    if (exclusive_rec_selection) fl_set_button(obj,1);
  alternate_app_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,255,140,220,22,"enable alternate appearance");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
    if (alternate_appearance) fl_set_button(obj,1);

  obj = fl_add_labelframe(FL_ENGRAVED_FRAME,250,180,240,40,"Update");
  active_update_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,255,190,220,22,"enable active volume update");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
    if (active_update) fl_set_button(obj,1);

  obj = fl_add_labelframe(FL_ENGRAVED_FRAME,250,230,240,115,"Save");
  save_volumes_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,255,290,220,22,"save actual volumes");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
  save_recsrc_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,255,315,220,22,"save recording sources");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
  save_channel_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,255,240,220,22,"save channel settings");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
    fl_set_button(obj,1);
  save_app_button=obj=fl_add_checkbutton(FL_PUSH_BUTTON,255,265,220,22,"save appearance");
    fl_set_object_callback(obj, mix_config_dummy_cb, 0);
    fl_set_button(obj,1);

  obj = fl_add_button(FL_NORMAL_BUTTON,250,370+addy,70,25,"Save");
    fl_set_object_callback(obj, mix_config_save_cb, 0);
    fl_set_object_shortcut(obj, "#s", TRUE);
  obj = fl_add_button(FL_NORMAL_BUTTON,335,370+addy,70,25,"Apply");
    fl_set_object_callback(obj, mix_config_apply_cb, 0);
    fl_set_object_shortcut(obj, "#a", TRUE);
  obj = fl_add_button(FL_NORMAL_BUTTON,420,370+addy,70,25,"Close");
    fl_set_object_callback(obj, mix_config_close_cb, 0);
    fl_set_object_shortcut(obj, "#c", TRUE);

  fl_end_form();
  fl_show_form(config_form, FL_PLACE_MOUSE, FL_TRANSIENT, PRG_NAME " config");
}

/*----------------------------------------------------------------------------------*/
/* options functions */
/*----------------------------------------------------------------------------------*/

void print_usage(char *argv[], int long_help)
{
  printf("\n%s %s by Sven Anders\n\n", PRG_NAME, VERSION);
  printf("usage: %s [-i] [-p] [-v] [-q] [-h] [-?]\n", argv[0]);
  printf("            [-f [<config filename>]] [-d <mixer device>]\n");
  printf("            [-l [<channel>]] [-m [<channel>]] [-r [<channel>]]\n");
  printf("            [-s [<channels>] [<mono/both> | <left> <right>]]\n\n");
  printf("\t-h displays this help message\n");
  printf("\t-? displays a verbose help message\n");

  if (long_help)
  {
    printf("\t-i starts in inactive mode.\n");
    printf("\t-v turns verbose messages on.\n");
    printf("\t-p print all channel's values.\n");
    printf("\t-q run without X, (quits after setting the volumes).\n");
    printf("\n");
    printf("\t-d specify another mixer device.\n");
    printf("\t-f load configuration from file <filename>.\n");
    printf("\t   (if <filename> is omitted or '-s' and '-q' options are\n");
    printf("\t    specified, no configuration file will be loaded).\n");
    printf("\n");
    printf("\t-l lock (all) channel(s).\n");
    printf("\t-m mute (all) channel(s).\n");
    printf("\t-r set (all) channel(s) to record.\n");

    printf("\t-s set <channels> to <volume specifier>.\n");
    printf("\t   <channels> is:\n");
    printf("\t      <channel>[[,|-][<channels>]]\n");
    printf("\t   <channel> is:\n");
    printf("\t      between 0 and %u\n", mix.devs);
    printf("\t   <volume specifier> is:\n");
    printf("\t      [<both> | <left> <right>]  for stereo channels\n");
    printf("\t      [<mono>]  for mono channels\n");
    printf("\t   <mono>, <both>, <left> and <right> are:\n");
    printf("\t      =<value>  set volume to <value>\n");
    printf("\t      +<value>  increase volume by <value>\n");
    printf("\t      -<value>  decrease volume by <value>\n");
    printf("\t   <value> is:\n");
    printf("\t       between 0 and %u\n", MAX_MIXER_VALUE);
    printf("\n");
  }

  exit(0);
}

typedef enum { UNKNOWN, NONE, CHANNEL, SET_VOLUME, INC_VOLUME, DEC_VOLUME } value_type;
typedef enum { NONEC, ONE, SEVERAL, ALL } channel_type;

typedef struct 
{
  char          *string;
  value_type    type;
  int           value;
} option_type;

void set_new_volume(int channel, int value_left, value_type type_left, 
		                 int value_right, value_type type_right)
{
  if ((channel < 0) || (channel > mix.devs))
  {
    printf("error: channel %u ist out of range\n", channel);
    return;
  }

  if (type_left == INC_VOLUME)  value_left = mchannel[channel].left + value_left;
  if (type_right == INC_VOLUME) value_right = mchannel[channel].right + value_right;
  if (type_left == DEC_VOLUME)  value_left = mchannel[channel].left - value_left;
  if (type_right == DEC_VOLUME) value_right = mchannel[channel].right - value_right;

  if (value_left < 0) value_left = 0;
  if (value_left > MAX_MIXER_VALUE) value_left = MAX_MIXER_VALUE;
  if (value_right < 0) value_right = 0; 
  if (value_right > MAX_MIXER_VALUE) value_right = MAX_MIXER_VALUE;

  if (verbose)
  {
    printf("setting volume of channel %i, to %i", channel, value_left);
    if (mchannel[channel].stereo) printf(" / %i.", value_right);
    printf("\n");
  }
  set_volume(channel, value_left, value_right);
}

void resolve_options(int argc, char *argv[])
{
  int i, j, k;
  int skip, passed, priority;
  int skip_config_load,
      set_volume_option,
      quit_option;
  option_type *option;
  channel_type type_channel;
  int value_left, value_right;
  value_type type_left, type_right;
  char *pos1, *pos2;
  char sc;

  /* first pass */

  skip_config_load = FALSE;
  set_volume_option = FALSE;
  quit_option = FALSE;
  option = (option_type *)malloc(sizeof(option_type)*argc);

  for (i=1; i<argc; i++)
  {
    option[i].string = argv[i];
    option[i].type = UNKNOWN;
    option[i].value = 0;

    if ((argv[i][0] == '-') && (strlen(argv[i])>1))
    {
      if (argv[i][1] == 's') set_volume_option = TRUE;
      if (argv[i][1] == 'q') quit_option = TRUE;
      if ((argv[i][1] >= '0') && (argv[i][1] <= '9'))
      {
        option[i].type = DEC_VOLUME;
        option[i].value = atoi(argv[i]+1);   
      }
      else
        option[i].type = NONE;
    }
    if (argv[i][0] == '+')
    {
      option[i].type = INC_VOLUME;
      option[i].value = atoi(argv[i]+1);   
    }
    if (argv[i][0] == '=')
    {
      option[i].type = SET_VOLUME;
      option[i].value = atoi(argv[i]+1);   
    }
    if ((argv[i][0] >= '0') && (argv[i][0] <= '9'))
    {
      option[i].type = CHANNEL;
      option[i].value = atoi(argv[i]);
    }
  }
  if (quit_option && set_volume_option) skip_config_load = TRUE;

  /* second pass */

  i = 1; /* 0 ist programname */
  passed = 1;
  priority = 0;
  while (priority < 10)
  {
    if ( (priority == 4) ||
	 ((i < argc) && (argv[i][0] == '-') && (argv[i][1] != '\0')) )
    {
      switch (priority)
      {
        case 0: switch (argv[i][1])
	        {
		   case 'h': 
		   case 'H': argv[i][0]='\0';
		             passed++;
		             print_usage(argv, FALSE);
			     break;

		   case 'v': argv[i][0]='\0';
		             passed++;
		             verbose = TRUE;
			     break;

                   default: break;
		} /*switch*/
	        break;

        case 1: switch (argv[i][1])
	        {
                  case 'd': argv[i][0]='\0';
		            passed++;
		            if ((i+1 < argc) && (option[i+1].type != NONE))
		            {
			      dev_name = (char *)malloc(strlen(option[i+1].string));
			      strcpy(dev_name, option[i+1].string);
			      if (verbose) printf("set device to: %s\n", dev_name);
			      /* initialize mixer struct for new mixer */
			      init_channel_struct();
			      i++; passed++;
			      argv[i][0]='\0';
			    }
		            else
			      fprintf(stderr,"error: no mixer device specified!\n");
                            break;

                  default: break;
		} /*switch*/
	        break;

        case 2:	switch (argv[i][1])
	        {
		  case 'p': argv[i][0]='\0';
		            passed++;
		            print_values();
			    break;

                  default: break;
		} /*switch*/
                break;

        case 3: switch (argv[i][1])
	        {
		  case 'f': argv[i][0]='\0';
		            passed++;
		            if ((i+1 < argc) && (option[i+1].type != NONE))
			    {
			      config_file = (char *)malloc(strlen(option[i+1].string));
			      strcpy(config_file, option[i+1].string);
			      if (verbose) printf("new configuration file: %s\n", config_file);
			      i++; passed++;
			      argv[i][0]='\0';
			    }
			    else
			    {
			      skip_config_load = TRUE;
			      if (verbose) printf("skipping loading of configuration file.\n");
			    }
			    break;

                  default: break;
		} /*switch*/
	        break;

        case 4: if (i==1) /*<- only load once) */
	        {
		  if (skip_config_load)
		  {
		    if (verbose) printf("skipping loading of configuration file.\n");
		  }
		  else
		    mix_config_load();
		}
	        break;

        case 5: switch (argv[i][1])
	        {
		  case 's': passed++;
		            value_left = MAX_MIXER_VALUE;
		            type_left = SET_VOLUME;
			    value_right = MAX_MIXER_VALUE;
			    type_right = SET_VOLUME;
			    type_channel = NONEC;
			    skip = 0;

			    if ((i+1 < argc) &&
				(option[i+1].type != UNKNOWN) &&
				(option[i+1].type != NONE))
			    {
			      /* Is 1. option an volume value ? */
			      if ((option[i+1].type == INC_VOLUME) ||
				  (option[i+1].type == DEC_VOLUME) ||
				  (option[i+1].type == SET_VOLUME))
			      {
				skip++;
				type_channel = ALL;
				type_left = option[i+1].type;
				value_left = option[i+1].value;

				/* Is 2. option an value too ? (for stereo) */
				if ((i+2<argc) && 
				    ((option[i+2].type == INC_VOLUME) ||
				     (option[i+2].type == DEC_VOLUME) ||
				     (option[i+2].type == SET_VOLUME)))
				{
				  skip++;
				  type_right = option[i+2].type;
				  value_right = option[i+2].value;
				}
				else
				{
				  type_right = type_left;
				  value_right = value_left;
				}
			      }

			      /* Is 1. option an channel (or channels ?) */
			      if (option[i+1].type == CHANNEL)
			      {
				skip++;
				if ((strchr(option[i+1].string,'-') != NULL) ||
				    (strchr(option[i+1].string,',') != NULL))
				  type_channel = SEVERAL;
				else
				  type_channel = ONE;

				/* Is 2. option an value ? */
				if ((i+2<argc) && 
				    ((option[i+2].type == INC_VOLUME) ||
				     (option[i+2].type == DEC_VOLUME) ||
				     (option[i+2].type == SET_VOLUME)))
				{
				  skip++;
				  type_left = option[i+2].type;
				  value_left = option[i+2].value;

				  /* Is 3. option an value too ? (for stereo) */
				  if ((i+3<argc) && 
				      ((option[i+3].type == INC_VOLUME) ||
				       (option[i+3].type == DEC_VOLUME) ||
				       (option[i+3].type == SET_VOLUME)))
				  {
				    skip++;
				    type_right = option[i+3].type;
				    value_right = option[i+3].value;
				    if ((type_channel == ONE) &&
					(!mchannel[option[i+1].value].stereo))
				    {
				      fprintf(stderr, "Useless value %s. Channel %u is mono.\n", 
					      option[i+3].string, option[i+1].value);
				    }
				  }
				  else
				  {
				    type_right = type_left;
				    value_right = value_left;
				  }
				}
			      }
			    }

			    if (type_channel == NONEC)
			    {
			      fprintf(stderr,"At least a overall volume setting is required.\n");
			    }
			    else
			    {
			      /* set volume */

			      if (type_channel == ONE)
			      {
				set_new_volume(option[i+1].value, 
					       value_left, type_left,
					       value_right, type_right);
			      }
			      else
			      {
				if (type_channel == ALL)
				  {
				    for (j=0;j<mix.devs;j++)
				      set_new_volume(j, value_left, type_left,
						     value_right, type_right);
				  }
			        else
				{
				  if (type_channel == SEVERAL)
				  {
				    pos1 = option[i+1].string;
				    pos2 = option[i+1].string;
				    do
				    {
				      pos2++;
				      if ((*pos2 == ',') || (*pos2 == '\0'))
				      {
					sc=*pos2; *pos2='\0'; j=atoi(pos1); *pos2=sc;
					set_new_volume(j, value_left, type_left,
						       value_right, type_right);
					pos1 = pos2+1;
				      }
				      else
				      {
					if (*pos2 == '-')
					{
					  sc=*pos2; *pos2='\0'; j=atoi(pos1); *pos2=sc;
					  pos2++; pos1 = pos2;
					  if ((*pos1 == ',') || (*pos1 == '\0'))
					    k=mix.devs;
					  else
					  {
					    pos2 = pos1;
					    while ((*pos2 != '\0') && (*pos2 != ',')) pos2++;
					    sc=*pos2; *pos2='\0'; k=atoi(pos1); *pos2=sc;
					  }
					  for (;j<=k;j++)
					    set_new_volume(j, value_left, type_left,
							   value_right, type_right);

					  pos1 = pos2+1;
					}
				      } /*else*/
				    } while (*pos2 != '\0');
				  } /*if*/
				} /*else*/
			      } /*else*/
			    } /*else*/
			    passed = passed + skip;
			    for (j=i; j<i+skip+1; j++) argv[i][0]='\0';
			    i = i + skip;
			    break;

                  default: break;
		} /*switch*/
	        break;

        case 6: switch (argv[i][1])
	        {
		  case 'l': argv[i][0]='\0'; 
		            passed++;
		            if ((i+1<argc) && (option[i+1].type == CHANNEL))
			    {
			      if (verbose) printf("locking channel %u...\n", option[i+1].value);
			      un_lock_channel(option[i+1].value);
			      i++; passed++;
			      argv[i][0]='\0'; 
			    }
			    else
			    {
			      if (verbose) printf("locking all channels...\n");
			      for (j=0;j<mix.devs;j++) un_lock_channel(j);
			    }
			    break;

		  case 'm': argv[i][0]='\0'; 
		            passed++;
		            if ((i+1<argc) && (option[i+1].type == CHANNEL))
			    {
			      if (verbose) printf("muting channel %u...\n", option[i+1].value);
			      un_mute_channel(option[i+1].value);
			      i++; passed++;
			      argv[i][0]='\0'; 
			    }
			    else
			    {
			      if (verbose) printf("muting all channels...\n");
			      for (j=0;j<mix.devs;j++) un_mute_channel(j);
			    }
			    break;
                 
		  case 'r': argv[i][0]='\0'; 
		            passed++;
		            if ((i+1<argc) && (option[i+1].type == CHANNEL))
			    {
			      if (verbose) printf("setting channel %u to record source...\n", option[i+1].value);
			      un_rec_channel(option[i+1].value);
			      i++; passed++;
			      argv[i][0]='\0'; 
			    }
			    else
			    {
			      if (verbose) printf("setting all channel to record sources...\n");
			      for (j=0;j<mix.devs;j++) un_rec_channel(j);
			    }
			    break;

		  default: break;
		} /*switch*/
                break;

        case 7: switch (argv[i][1])
	        {
		  case 'i': argv[i][0]='\0';
                    	    passed++;
		            active_update = FALSE;
			    break;
		  default: break;
		} /*switch*/
	        break;

        case 8:	switch (argv[i][1])
	        {
		  case '?': argv[i][0]='\0';
		            passed++;
		            print_usage(argv, TRUE);
			    break;

		  case 'q': free(option);
		            exit(0);
		            break;

		  default: break;
		} /*switch*/
	        break;

        case 9: if (option[i].type == UNKNOWN)
	        {
                  fprintf(stderr, "unknown option %s\n", option[i].string);
                }
                passed++;
                break;

        default: exit(0); break; /*should never happen...*/

      } /*switch (priority)*/

    } /*if*/

    if (i < argc-1)
      i++;
    else
    {
      i=1; priority++;
    }

  } /*while (passed)*/

  free(option);

} /*function*/

/*------------------------------------------------------------------------------------*/
/* main program */
/*------------------------------------------------------------------------------------*/

int main(int argc, char *argv[])
{
  FL_IOPT fl_defaults;

  /* initialize mixer struct */
  init_channel_struct();

  /* resolve option */
  resolve_options(argc, argv);

  /* set some new xforms defaults */
  fl_defaults.borderWidth = -2;
  fl_set_defaults(FL_PDBorderWidth, &fl_defaults);
 
  /* initialize forms library */
  fl_initialize(&argc, argv, PRG_NAME, 0, 0);

  /* create main window form */   
  create_form_mix();

  /* set alarm handler (for active update) */
  /* Attention: fl_msleep does not work any more ! */
  /*            use 'my_msleep' instead ! */
  sig_alarm_action.sa_handler = sig_alarm_handler;
  sig_alarm_action.sa_flags = SA_NOMASK;
  #ifndef alpha
    sig_alarm_action.sa_restorer = NULL;
  #endif
  sigaction(SIGALRM, &sig_alarm_action, &old_sig_alarm_action);
  setitimer(ITIMER_REAL, &sig_alarm_delay, &old_sig_alarm_delay);

  /* start xforms loop */
  fl_do_forms();

  /* exit with return code */
  return 0;
}
